# from importlib.metadata import version
# TODO replace * with more expliced exports for the final version of the package 
# # from .radius_search import *
# import utils
# import testing
# import illustrations
# import radius_search
from . import main
# from .main import *
# from .radius_search.optimal_grid_spacing import *
# from .radius_search.offset_regions import *
# from ..utils.general import *
# from ..illustrations.illustrate_nested_grid import *
# from ..illustrations.illustrate_optimal_grid_spacing import *
# from ..illustrations.illustrate_point_to_cell_region_assignment import *
# # from .nested_search import *
# from .radius_search.radius_search_class import *
# from ..illustrations.plot_utils import *
# from .radius_search.two_dimensional_weak_ordering_class import *
# from .main import *
# from ..illustrations.plot_utils import *
# TODO ensure to no exports imports (like np.array)?

# __version__ = version('primelocations')



from numpy import (
    array as _np_array, 
    linspace as _np_linspace,
    stack as _np_stack,
    arange as _np_arange, 
    unique, invert, flip, transpose, concatenate, sign, zeros, 
    min, max, equal, where, logical_or, logical_and, all, newaxis
)
# from numpy.linalg import norm
# from numpy.random import randint, random
from pandas import DataFrame as _pd_DataFrame
from math import log10 as _math_log10#,ceil,asin,acos
from aabpl.utils.general import ( flatten_list, time_function, visualize, depth )
from aabpl.illustrations.illustrate_optimal_grid_spacing import ( create_optimal_grid_spacing_gif, )
from aabpl.illustrations.plot_utils import map_2D_to_rgb, get_2D_rgb_colobar_kwargs
from aabpl.utils.distances_to_cell import ( get_always_contained_potentially_overlapped_cells, )
# from .nested_search import (aggregate_point_data_to_nested_cells, aggreagate_point_data_to_disks_vectorized_nested)
from .radius_search_class import (
    aggregate_point_data_to_cells,
    assign_points_to_cell_regions,
    aggreagate_point_data_to_disks_vectorized
)
from aabpl.valid_area import disk_cell_intersection_area
# from .radius_search.optimal_grid_spacing import (select_optimal_grid_spacing,)
from aabpl.testing.test_performance import time_func_perf, func_timer_dict
import matplotlib.pyplot as plt


class Bounds(object):
    __slots__ = ('xmin', 'xmax', 'ymin', 'ymax', 'np_array_of_bounds') # use this syntax to save some memory. also only create vars that are really neccessary
    def __init__(self, xmin:float, xmax:float, ymin:float, ymax:float):
        self.xmin = xmin
        self.xmax = xmax
        self.ymin = ymin
        self.ymax = ymax
    #
#

# TODO potentially add nestes GridCell classtype the 
class GridCell(object):
    # __slots__ = ('id', 'row_col_nr', ...) # use this syntax to save some memory. also only create vars that are really neccessary
    def __init__(self, row_nr, col_nr, y_steps, x_steps):
        self.id = (row_nr,col_nr)
        self.centroid = (y_steps[row_nr:row_nr+2].sum()/2,x_steps[col_nr:col_nr+2].sum()/2)
        self.bounds = Bounds(xmin=x_steps[col_nr], ymin=y_steps[row_nr], xmax=x_steps[col_nr+1], ymax=y_steps[row_nr+1])
        self.pt_ids = []
        self.excluded = None

    def add_pt_id(self, pt_id):
        self.pt_ids = [*self.pt_ids, pt_id]
        
    def add_pt_ids(self, pt_ids):
        self.pt_ids = [*self.pt_ids, *pt_ids]
    
    def set_excluded_area(self,excluded_area):
        pass

    def make_sparse(self):
        # self.pt_ids # make _np_array or tuple
        # tighten bounds
        pass            
    def tighten_bounds(self, pt_ids):
        self.pt_ids = [*self.pt_ids, *pt_ids]
    
    def plot(self, facecolor, edgecolor, ):
        pass
#


class Grid(object):
    """

    """
    @time_func_perf
    def __init__(
        self,
        xmin:float,
        xmax:float,
        ymin:float,
        ymax:float,
        set_fixed_spacing:float=None,
        radius:float=750,
        n_points:int=10000,
        silent = False,
        ):

        """

        """
        if set_fixed_spacing:
            spacing = set_fixed_spacing
        else:
            # find optimal spacing TODO
            print("TODO find optimal spacing for",radius, n_points)
            spacing = 1.
        self.spacing = spacing

        # TODO total_bounds should also contain excluded area if not contained 
        # min(points.total_bounds+radius, max(points.total_bounds, excluded_area_total_bound))  
      
        x_padding = ((xmin-xmax) % spacing)/2
        y_padding = ((ymin-ymax) % spacing)/2
        
        self.n_x_steps = n_x_steps = -int((xmin-xmax)/spacing) # round up
        self.n_y_steps = n_y_steps = -int((ymin-ymax)/spacing) # round up 
        self.total_bounds = total_bounds = Bounds(xmin=xmin-x_padding,xmax=xmax+x_padding,ymin=ymin-y_padding,ymax=ymax+y_padding)
        
        self.x_steps = x_steps = _np_linspace(total_bounds.xmin, total_bounds.xmax, n_x_steps)
        # self.y_steps = y_steps = _np_linspace(total_bounds.ymax, total_bounds.ymin, n_y_steps)
        self.y_steps = y_steps = _np_linspace(total_bounds.ymin, total_bounds.ymax, n_y_steps)
        
        self.id_y_mult = id_y_mult = 10**(int(_math_log10(n_x_steps))+1)
        
        self.row_ids = _np_arange(n_y_steps-1)
        self.col_ids = _np_arange(n_x_steps-1)
        
        self.ids = tuple(flatten_list([[(row_id, col_id) for col_id in range(n_x_steps-1)] for row_id in range(n_y_steps-1)]))
        class CellDict(dict):
            def __init__(self, x_steps, y_steps, id_y_mult):
                self.x_steps = x_steps
                self.y_steps = y_steps
                self.id_y_mult = id_y_mult

            def id_to_row_col(self,id:int): 
                return (id // self.id_y_mult, id % self.id_y_mult)
            
            def row_col_to_id(self,row_nr:int,col_nr:int): 
                return row_nr * self.id_y_mult + col_nr
            
            def add_new(self,row,col):
                setattr(self, str(self.row_col_to_id(row,col)), GridCell(
                    row, col, self.x_steps, self.y_steps
                ))
            
            def get_by_row_col(self,row,col):
                return getattr(self, str(self.row_col_to_id(row,col)))
            
            def get_by_id(self,id):
                return getattr(self, str(id))
            
            def get_or_create(self,row,col):
                id = str(self.row_col_to_id(row,col))
                if not hasattr(self, id):
                    self.add_new(row,col)
                return self.get_by_id(id)
            
            def add_pts(self, pts, row, col):
                cell = self.get_or_create(row,col)
                cell.add_pts(pts)
            
            def add_pt(self, pt, row, col):
                cell = self.get_or_create(row,col)
                cell.add_pt(pt)
            
        self.row_col_stack = _np_stack([
                    _np_array([[row_id for col_id in range(n_x_steps-1)] for row_id in range(n_y_steps-1)]).flatten(),
                    _np_array([[col_id for col_id in range(n_x_steps-1)] for row_id in range(n_y_steps-1)]).flatten(),
                ])
        self.cells = CellDict(x_steps=x_steps, y_steps=y_steps, id_y_mult=id_y_mult,)

        self.cells = _np_array([[
            GridCell(row_id, col_id, y_steps=y_steps, x_steps=x_steps) for col_id in range(n_x_steps-1)
            ] for row_id in range(n_y_steps-1)]).flatten()

        self.row_col_to_centroid = {g_row_col:centroid for (g_row_col,centroid) in flatten_list([
                [((row_id,col_id),(x_steps[col_id:col_id+2].mean(), y_steps[row_id:row_id+2].mean())) for col_id in range(n_x_steps-1)] 
                for row_id in range(n_y_steps-1)]
                )}
        self.centroids = _np_array([centroid for centroid in flatten_list([
                [(x_steps[col_id:col_id+2].mean(), y_steps[row_id:row_id+2].mean()) for col_id in range(n_x_steps-1)] 
                for row_id in range(n_y_steps-1)]
                )])
        self.id_to_bounds = {row_id*id_y_mult+col_id: bounds for ((row_id,col_id),bounds) in flatten_list([
                [((row_id,col_id),((x_steps[col_id], y_steps[row_id]), (x_steps[col_id+1], y_steps[row_id+1]))) for col_id in range(n_x_steps-1)] 
                for row_id in range(n_y_steps-1)]
                )}
        self.bounds = flatten_list([
                [((x_steps[col_id], y_steps[row_id]), (x_steps[col_id+1], y_steps[row_id+1])) for col_id in range(n_x_steps-1)] 
                for row_id in range(n_y_steps-1)])
        self.cell_dict = dict()
        if not silent:
            print('Create grid with '+str(n_y_steps-1)+'x'+str(n_x_steps-1)+'='+str((n_y_steps-1)*(n_x_steps-1)))
        #
    #
    # add functions
    aggregate_point_data_to_cells = aggregate_point_data_to_cells
    # aggregate_point_data_to_nested_cells = aggregate_point_data_to_nested_cells
    assign_points_to_cell_regions = assign_points_to_cell_regions
    aggreagate_point_data_to_disks_vectorized = aggreagate_point_data_to_disks_vectorized
    # aggreagate_point_data_to_disks_vectorized_nested = aggreagate_point_data_to_disks_vectorized_nested
    # append a variable to pts_df that indicates the share of valid area float[0,1] 
    disk_cell_intersection_area = disk_cell_intersection_area
    def create_sparse_grid(self):
        
        
        return
        # TODO shall this be another class? A new instane of the grid? Or create a sparse copy within itself? Or overwrite itself with an own c?
        print("Make this grid sparse")
    def create_nested_grid(self):
        print("Create nested grid")
    #

    def plot_grid(self, fig=None, ax=None,):
        if ax is None:
            fig, ax = plt.subplots(nrows=3, figsize=(15,25))
            imshow_kwargs = {
                'xmin':self.x_steps.min(),
                'ymin':self.y_steps.min(),
                'xmax':self.x_steps.max(),
                'ymax':self.y_steps.max(),
            }
            extent=[imshow_kwargs['xmin'],imshow_kwargs['xmax'],imshow_kwargs['ymax'],imshow_kwargs['ymin']]
            X = _np_array([[map_2D_to_rgb(x,y, **imshow_kwargs) for x in  self.x_steps[:-1]] for y in self.y_steps[:-1]])
            # ax.flat[0].imshow(X=X, interpolation='none', extent=extent)
            # ax.flat[0].pcolormesh([self.x_steps, self.y_steps], X)
            ax.flat[0].pcolormesh(X, edgecolor="black", linewidth=1/max([self.n_x_steps, self.n_y_steps])/1.35)
            # ax.flat[0].set_aspect(2)
            colorbar_kwargs = get_2D_rgb_colobar_kwargs(**imshow_kwargs)
            cb = plt.colorbar(**colorbar_kwargs[2], ax=ax.flat[0])
            cb.ax.set_xlabel("diagonal")
            cb = plt.colorbar(**colorbar_kwargs[0], ax=ax.flat[0])
            cb.ax.set_xlabel("x/lon")
            cb = plt.colorbar(**colorbar_kwargs[1], ax=ax.flat[0])
            cb.ax.set_xlabel("y/lat") 
            ax.flat[0].set_xlabel('x/lon') 
            ax.flat[0].set_ylabel('y/lat') 
            ax.flat[0].title.set_text("Grid lat / lon coordinates")
            # ax.flat[0].set_xticks(self.x_steps, minor=True)
            # ax.flat[0].set_yticks(self.y_steps, minor=True)
            # ax.flat[0].grid(which='minor', color='w', linestyle='-', linewidth=0.002)

            imshow_kwargs = {
                'xmin':self.col_ids.min(),
                'ymin':self.row_ids.min(),
                'xmax':self.col_ids.max(),
                'ymax':self.row_ids.max(),
            }
            extent=[imshow_kwargs['xmin'],imshow_kwargs['xmax'],imshow_kwargs['ymax'],imshow_kwargs['ymin']]

            X = _np_array([[map_2D_to_rgb(x,y, **imshow_kwargs) for x in  self.col_ids] for y in self.row_ids])
            ax.flat[1].imshow(X=X, interpolation='none', extent=extent)
            # ax.flat[1].set_aspect(2)
            colorbar_kwargs = get_2D_rgb_colobar_kwargs(**imshow_kwargs)
            cb = plt.colorbar(**colorbar_kwargs[2], ax=ax.flat[1])
            cb.ax.set_xlabel("diagonal")
            cb = plt.colorbar(**colorbar_kwargs[0], ax=ax.flat[1])
            cb.ax.set_xlabel("col nr")
            cb = plt.colorbar(**colorbar_kwargs[1], ax=ax.flat[1])
            cb.ax.set_xlabel("row nr") 
            ax.flat[1].set_xlabel('row nr') 
            ax.flat[1].set_ylabel('col nr') 
            ax.flat[1].title.set_text("Grid row / col indices")
            # ax.flat[1].set_xticks(self.col_ids, minor=True)
            # ax.flat[1].set_yticks(self.row_ids, minor=True)
            # ax.flat[1].grid(which='minor', color='black', linestyle='-', linewidth=0.003)
            
            X = _np_array([[len(self.id_to_pt_ids[(row_id, col_id)]) if (row_id, col_id) in self.id_to_pt_ids else 0 for col_id in self.col_ids] for row_id in self.row_ids])
            p = ax.flat[2].pcolormesh(X, cmap='Reds')
            ax.flat[2].set_xlabel('row nr') 
            ax.flat[2].set_ylabel('col nr') 
            plt.colorbar(p)
#

class ExludedArea:
    def __init__(self,excluded_area_geometry_or_list, grid:Grid):
        # recursively split exluded area geometry along grid 
        # then sort it into grid cell
        
        pass
#





from numpy import (
    array as _np_array,
    ndarray as _np_ndarray,
    arange as _np_arange,
    hstack as _np_hstack,
    zeros as _np_zeros,
    unique as _np_unique, 
    concatenate as _np_concatenate,
    equal as _np_equal, 
    logical_or as _np_logical_or, 
    all as _np_all, 
)
from numpy.linalg import norm as _np_linalg_norm
from pandas import (DataFrame as _pd_DataFrame, cut as _pd_cut, concat as _pd_concat) 
from aabpl.utils.general import (   flatten_list, DataFrameRelation,arr_to_tpls )
from aabpl.illustrations.illustrate_point_to_disk import (illustrate_point_disk)
from aabpl.illustrations.plot_pt_vars import create_plots_for_vars
from aabpl.testing.test_performance import time_func_perf, func_timer_dict

################ aggreagate_point_data_to_disks_vectorized ######################################################################################
@time_func_perf
def aggreagate_point_data_to_disks_vectorized(
    grid:dict,
    pts_df_search_source:_pd_DataFrame,
    pts_df_search_target:_pd_DataFrame=None,
    radius:float=0.0075,
    sum_names:list=['employment'],
    y_coord_name:str='lat',
    x_coord_name:str='lon',
    row_name:str='id_y',
    col_name:str='id_x',
    cell_region_name:str='cell_region',
    sum_suffix:str='_750m',
    exclude_pt_itself:bool=True,
    plot_radius_sums:dict=None,
    plot_pt_disk:dict=None,
    silent = False,
):
    """
    
    """
    
    if pts_df_search_target is None:
        pts_df_search_target = pts_df_search_source 
    
    # unpack grid_data 
    grid_id_to_pt_ids = grid.id_to_pt_ids
    grid_id_to_sums = grid.id_to_sums
    pt_id_to_row_col = grid.pt_id_to_row_col
    sparse_grid_ids = set(grid_id_to_pt_ids)
    
    region_id_to_contained_cells = grid.search.region_id_to_contained_cells
    region_id_to_overlapped_cells = grid.search.region_id_to_overlapped_cells
    cells_contained_in_all_disks = grid.search.cells_contained_in_all_disks

    id_y_mult=grid.id_y_mult
    pt_id_to_xy_coords = grid.search.target.pt_id_to_xy_coords
    pt_id_to_vals = grid.search.target.pt_id_to_vals
    n_pts = len(pts_df_search_source)

    # initialize columns and/or reset to zero 
    sum_radius_names = [(cname+sum_suffix) for cname in sum_names]
    pts_df_search_source[sum_radius_names] = 0
    
  
    sums_within_disks = _np_zeros((n_pts, len(sum_names)))
    
    if plot_pt_disk is not None:
        if not 'pt_id' in plot_pt_disk:
            plot_pt_disk['pt_id'] = pts_df_search_source.index[int(n_pts//2)]
        
    zero_sums = _np_zeros(len(sum_names),dtype=int)
    # time_function(time_dict)#1
    pts_df_search_source['initial_sort'] = range(len(pts_df_search_source))
    pts_df_search_source.sort_values([row_name, col_name, cell_region_name], inplace=True)
    last_pt_row_col = (-1, -1)
    last_cell_region_id = -1
    counter_new_cell = 0
    counter_new_contain_region = 0
    counter_new_overlap_region = 0

    for (i, pt_id, a_pt_ycoord, a_pt_xcoord, (pt_row,pt_col), contain_region_id, overlap_region_id, cell_region_id) in zip(
        range(n_pts),
        pts_df_search_source.index,
        pts_df_search_source[y_coord_name].values, 
        pts_df_search_source[x_coord_name].values,
        pts_df_search_source[[row_name, col_name]].values,
        pts_df_search_source[cell_region_name].values // grid.search.contain_region_mult,
        pts_df_search_source[cell_region_name].values % grid.search.contain_region_mult,
        pts_df_search_source[cell_region_name].values,
        
        
        ):
        # (pt_row, pt_col) = pt_id_to_row_col[pt_id]
        
        # as pts are sorted by grid cell update only if grid cell changed
        if not (pt_row, pt_col) == last_pt_row_col:
            counter_new_cell += 1
            cells_cntd_by_pt_cell = sparse_grid_ids.intersection([(row+pt_row,col+pt_col) for row,col in cells_contained_in_all_disks])
            # cells_cntd_by_pt_cell = [cell_id for cell_id in (id_y_mult*(cells_contained_in_all_disks[:,0]+(pt_row))+(
            #     cells_contained_in_all_disks[:,1]+pt_col)) if cell_id in sparse_grid_ids] 
            sums_cells_cntd_by_pt_cell = (_np_array([grid_id_to_sums[g_id] for g_id in cells_cntd_by_pt_cell]).sum(axis=0) 
                                      if len(cells_cntd_by_pt_cell)>0 else zero_sums)
        #
            
        if not (pt_row, pt_col) == last_pt_row_col or last_contain_region_id != contain_region_id:
            counter_new_contain_region += 1
            # if cell changed or cell region changed
            # this can be improve to capture only changes of cell region conatain ids 
            cells_contained_by_pt_region = sparse_grid_ids.intersection([(row+pt_row,col+pt_col) for row,col in region_id_to_contained_cells[cell_region_id]])
            # cells_contained_by_pt_region = [cell_id for cell_id in (id_y_mult*(region_id_to_contained_cells[cell_region_id][:,0]+(pt_row))+(
            #     region_id_to_contained_cells[cell_region_id][:,1]+pt_col)) if cell_id in sparse_grid_ids]
                           
            sums_contained_by_pt_region = (_np_array([grid_id_to_sums[g_id] for g_id in cells_contained_by_pt_region]).sum(axis=0) 
                                             if len(cells_contained_by_pt_region)>0 else zero_sums)

        if not (pt_row, pt_col) == last_pt_row_col or last_overlap_region_id != overlap_region_id:
            counter_new_overlap_region += 1
            cells_overlapped_by_pt_region = sparse_grid_ids.intersection([(row+pt_row,col+pt_col) for row,col in region_id_to_overlapped_cells[cell_region_id]])
            # cells_overlapped_by_pt_region = [cell_id for cell_id in (id_y_mult*(region_id_to_overlapped_cells[cell_region_id][:,0]+(pt_row))+(
            #     region_id_to_overlapped_cells[cell_region_id][:,1]+pt_col)) if cell_id in sparse_grid_ids]
            
            pts_in_cells_overlapped_by_pt_region = _np_array(flatten_list([
                grid_id_to_pt_ids[cell_id] for cell_id in cells_overlapped_by_pt_region
            ]))
        #

        if len(pts_in_cells_overlapped_by_pt_region) > 0:
            pts_in_radius = pts_in_cells_overlapped_by_pt_region[(_np_linalg_norm( # create a mask of boolean values indicating if point are within radius 
                _np_array([pt_id_to_xy_coords[pt_id] for pt_id in pts_in_cells_overlapped_by_pt_region]) -
                _np_array([a_pt_xcoord, a_pt_ycoord]), 
            axis=1) <= radius)]
            
            overlapping_cells_sums = _np_array([pt_id_to_vals[pt_id] for pt_id in pts_in_radius]).sum(axis=0) if len(pts_in_radius) > 0 else zero_sums
        else:
            # else no points in radius thus return vector of _np_zeros
            overlapping_cells_sums = zero_sums

        # combine sums from three steps.
        sums_full_disk = sums_cells_cntd_by_pt_cell + sums_contained_by_pt_region + overlapping_cells_sums
        
        # append result 
        sums_within_disks[i,:] = sums_full_disk
        
        if plot_pt_disk is not None and pt_id == plot_pt_disk['pt_id']:
            illustrate_point_disk(
                grid=grid,
                pts_df_search_source=pts_df_search_source,
                pts_df_search_target=pts_df_search_target,
                radius=radius,
                sum_names=sum_names,
                x_coord_name=x_coord_name,
                y_coord_name=y_coord_name,
                cells_cntd_by_pt_cell=[(row+pt_row,col+pt_col) for row,col in cells_contained_in_all_disks],
                cells_contained_by_pt_region=[(row+pt_row,col+pt_col) for row,col in region_id_to_contained_cells[cell_region_id]],
                cells_overlapped_by_pt_region=[(row+pt_row,col+pt_col) for row,col in region_id_to_overlapped_cells[cell_region_id]],
                pts_in_cell_contained_by_pt_region=_np_array(flatten_list([
                        grid_id_to_pt_ids[cell_id] for cell_id in cells_cntd_by_pt_cell
                    ]+[
                        grid_id_to_pt_ids[cell_id] for cell_id in cells_contained_by_pt_region
                    ])),
                pts_in_cells_overlapped_by_pt_region=pts_in_cells_overlapped_by_pt_region,
                pts_in_radius=pts_in_radius,
                **plot_pt_disk,
            )
        #

        # set id as last id for next iteration
        last_pt_row_col = (pt_row, pt_col)
        last_contain_region_id = contain_region_id
        last_overlap_region_id = overlap_region_id
    #
    pts_df_search_source[sum_radius_names] = pts_df_search_source[sum_radius_names].values + sums_within_disks
            
    if exclude_pt_itself and grid.search.tgt_df_contains_src_df:
        # substract data from point itself unless specified otherwise
        pts_df_search_source[sum_radius_names] = pts_df_search_source[sum_radius_names].values - pts_df_search_source[sum_names]
    print("counter_new_cell", round(counter_new_cell/len(pts_df_search_source),3), "cn",round(counter_new_contain_region/len(pts_df_search_source),3) , "ov",round(counter_new_overlap_region/len(pts_df_search_source),3) )
    
    if plot_radius_sums is not None:
        print('create plot for radius sums')
        create_plots_for_vars(
            grid=grid,
            colnames=_np_array([sum_names, sum_radius_names]),
            plot_kwargs=plot_radius_sums,
        )
    #
    pts_df_search_source.sort_values(['initial_sort'], inplace=True)
    pts_df_search_source.drop('initial_sort', axis=1, inplace=True)

    return pts_df_search_source[sum_radius_names]
#





from pandas import DataFrame as _pd_DataFrame
from numpy import (
    array as _np_array,
    linspace as _np_linspace,
    searchsorted as _np_searchsorted
)
from matplotlib.pyplot import (subplots as _plt_subplots)

def create_distribution_plot(
        sums_df:_pd_DataFrame,
        cluster_threshold_values:list,
        k_th_percentiles:list,
        radius=None,
        plot_kwargs:dict={},
):
    """
    TODO Descripiton
    """
    (n_random_points, ncols) = sums_df.shape
    # specify default plot kwargs and add defaults
    default_kwargs = {
        's':0.8,
        'color':'#eaa',

        'figsize': (5*ncols,5),
        'fig':None,
        'axs':None,
        
        'hlines':{'color':'red', 'linewidth':1},
        'vlines':{'color':'red', 'linewidth':1},
    }
    kwargs = {}
    for k in plot_kwargs:
        if k in [k for k,v in default_kwargs.items() if type(v)==dict]:
            kwargs[k] = {**default_kwargs.pop(k), **plot_kwargs.pop(k)}
    kwargs.update(default_kwargs)
    kwargs.update(plot_kwargs)
    figsize = kwargs.pop('figsize')
    fig = kwargs.pop('fig')
    axs = kwargs.pop('axs')

    if fig is None or axs is None:
        fig, axs = _plt_subplots(1,ncols, figsize=figsize)
    
    fig.suptitle(
        "Values for indicator" + ("" if ncols==1 else "s") + 
        ("" if radius is None else (" within "+ str(radius) +" distance")) +
        " around " + str(n_random_points)+ " randomly points drawn within valid area."
    )

    xmin, xmax = 0, 100
    xs = _np_linspace(xmin,xmax,n_random_points)
    vals = sums_df.values
    colnames = sums_df.columns

    for (i, colname, cluster_threshold_value, k_th_percentile) in zip(
        range(ncols), colnames, cluster_threshold_values, k_th_percentiles):
        ys = vals[:,i]
        ys.sort()
        ymin, ymax = ys.min(),ys.max()
        # round percentile value as far as necessary only 
        # e.g. threshold value is 10.5328... and next smaller/larger in distribution are 10.51..., 10.6... 
        # rounding to threshold value to firrst digit s.t. thath it lies between those value is sufficient (e.g. 10.53) 
        idx = _np_searchsorted(ys, cluster_threshold_value)
        next_smaller_val, next_larger_val = ys[max([0,idx-1])], ys[idx]
        sufficient_digits = next((
            i for i in range(100) if (
                (
                    (next_smaller_val == next_larger_val or cluster_threshold_values==next_smaller_val) and 
                    round(next_smaller_val,i)==next_smaller_val
                ) or (
                    round(next_larger_val, i) != round(cluster_threshold_value, i) and 
                    round(next_smaller_val, i) != round(cluster_threshold_value, i)
                )
        )),100)
        
        ax_title = (
            "Threshold value for "+str(k_th_percentile)+"th-percentile is "+
            str(round(cluster_threshold_value, sufficient_digits)) + " for "+ colname
        )
        
        # SELECT AX (IF MULTIPLE)
        ax = axs.flat[i] if ncols > 1 else axs
        
        # SET TITLE
        ax.set_title(ax_title)

        # SET TICKS
        xtick_steps, ytick_steps = 5, 5
        xticks = _np_array(sorted(
           [x for x in _np_linspace(xmin,xmax,xtick_steps) if abs(x-k_th_percentile) > (xmax-xmin)/(xtick_steps*2)] + 
           [k_th_percentile]
        ))
        ax.set_xticks(xticks, labels=xticks)
        yticks = _np_array(sorted([y for y in _np_linspace(ymin,ymax,ytick_steps) if abs(cluster_threshold_value-y)>(ymax-ymin)/(ytick_steps*10)] + [cluster_threshold_value]))
        ax.set_yticks(yticks, labels=[round(t, sufficient_digits) for t in yticks])

        # ADD CUTOFF LINES
        ax.hlines(y=cluster_threshold_value, xmin=xmin,xmax=xmax, **kwargs.pop('hlines'))
        ax.vlines(x=k_th_percentile, ymin=ymin,ymax=ymax, **kwargs.pop('vlines'))
        
        # ADD DISTRIUBTION PLOT
        ax.scatter(x=xs,y=ys, **plot_kwargs)
        
        # SET LIMITS
        ax.set_xlim([xmin,xmax])
        ax.set_ylim([ymin,ymax])
    #
#